/*
 * timer.c
 *
 */

#define MAIN_ENGINE_TIMER PD5
#define MAIN_ENGINE_SLEEP PA5
#define MAIN_ENGINE_DIRL  PA6
#define MAIN_ENGINE_DIRR  PA7

#define SENSOR_ENGINE_TIMER PD4
#define SENSOR_ENGINE_SLEEP PA3
#define SENSOR_ENGINE_DIR   PA4

volatile int _step_target;
volatile int _speed_hz;

// Set step speed in Hz (1Hz - 50kHz)
void SetEngineSpeedHz(int hz){

  int new_ICR1;
  
  // Determine new values for prescaler and TOP
  // Prescaler so that ~ 1000 < TOP < 65536

  if(hz == 0){
    // Stop timer
    TCCR1B = (TCCR1B & ~(0x7 << CS10));
    
    // Update current speed
    _speed_hz = 0;

    return;
  }
  else if(hz <= 50){
    // Prescaler 256
    TCCR1B = (TCCR1B & ~(0x7 << CS10)) | 
             (1 << CS12);

    new_ICR1 = F_CPU / (2 * 256 * hz);
  }
  else if(hz <= 200){
    // Prescaler 64
    TCCR1B = (TCCR1B & ~(0x7 << CS10)) | 
             (1 << CS11) | (1 << CS10);

    new_ICR1 = F_CPU / (2 * 64 * hz);
  }
  else if(hz <= 2000){
    // Prescaler 8
    TCCR1B = (TCCR1B & ~(0x7 << CS10)) | 
             (1 << CS11);

    new_ICR1 = F_CPU / (2 * 8 * hz);
  }
  else if(hz <= 50000){
    // Prescaler 1
    TCCR1B = (TCCR1B & ~(0x7 << CS10)) | 
             (1 << CS10);

    new_ICR1 = F_CPU / (2 * 1 * hz);
  }
  
  // Wait for the timer to reset before updating ICR1
  while(TCNT1)
    ;

  // Set new TOP
  ICR1 = new_ICR1;

  // Set fill rate 50%
  OCR1A = OCR1B = ICR1 >> 1;

  // Update current speed
  _speed_hz = hz;
}

// Initialize Engine Timer
// Timer won't start before prescaler is set
void InitEngineTimer(){

  // Reset registers
  TCCR1A = TCCR1B = 0;

  // Set OC1A and OC1B as output
  DDRD |= (1 << MAIN_ENGINE_TIMER) |
          (1 << SENSOR_ENGINE_TIMER);
  
  // Set PA5 (main engine) and 
  //     PA3 (sensor engine)
  // as output and enable sleep
  DDRA |= (1 << MAIN_ENGINE_SLEEP) | 
          (1 << SENSOR_ENGINE_SLEEP);

  PORTA &= ~((1 << MAIN_ENGINE_SLEEP) | 
             (1 << SENSOR_ENGINE_SLEEP));

  // Set PA7 and PA6 (main engine) and
  //     PA4 (sensor engine)
  // as output for direction control
  DDRA |= (1 << MAIN_ENGINE_DIRR) | 
          (1 << MAIN_ENGINE_DIRL) | 
          (1 << SENSOR_ENGINE_DIR);

  // Phase and Frequency correct PWM 16-bit
  // TOP in ICR1
  TCCR1B |= (1 << WGM13);

  // Enable timer output on OC1A
  // Set when upcounting clear when downcounting
  TCCR1A |= (1 << COM1A1) | (1 << COM1A0) |
            (1 << COM1B1) | (1 << COM1B0);

}

// Enable main engine
void EnableMainEngine(){

  // Enable interrupts on capture (TOP)
  TIMSK1 |= (1 << ICIE1);

  // Disable sleep
  PORTA |= (1 << MAIN_ENGINE_SLEEP);
  _delay_ms(10);

  // Set engines to stop and hold
  SetEngineSpeedHz(0);

}

// Disable main engine
void DisableMainEngine(){

  // Set engines to stop and hold
  SetEngineSpeedHz(0);

  // Disable interrupts on capture (TOP)
  TIMSK1 &= ~(1 << ICIE1);

  // Enable sleep
  PORTA &= ~(1 << MAIN_ENGINE_SLEEP);

}
// Enable sensor engine
void EnableSensorEngine(){

  // Enable interrupts on capture (TOP)
  TIMSK1 |= (1 << ICIE1);

  // Disable sleep
  PORTA |= (1 << SENSOR_ENGINE_SLEEP);
  _delay_ms(10);

  // Set engines to stop and hold
  SetEngineSpeedHz(0);

}

// Disable main engine
void DisableSensorEngine(){

  // Set engines to stop and hold
  SetEngineSpeedHz(0);

  // Disable interrupts on capture (TOP)
  TIMSK1 &= ~(1 << ICIE1);

  // Enable sleep
  PORTA &= ~(1 << SENSOR_ENGINE_SLEEP);

}

// Count total number of steps taken
ISR(TIMER1_CAPT_vect){

  _step_target--;

  if(_step_target <= 0){
    SetEngineSpeedHz(0);
    _delay_ms(10);
  }

}

void Move(int steps){

  int i;
  char str[64];

  EnableMainEngine();

  // Set direction on both engines as forward
  PORTA |=  (1 << MAIN_ENGINE_DIRL); 
  PORTA &= ~(1 << MAIN_ENGINE_DIRR);  

  _step_target = steps;

  for(i = 1; i <= 40; i++){
    SetEngineSpeedHz(i * 3);
    _delay_ms(20);
  }

}
